"""Dynamic logging context for CHIME project."""

import logging
import os
from logging import Logger
from typing import Literal, Optional

from pydantic import BaseModel, validator


class DynamicLoggerAdapter(logging.LoggerAdapter):
    """Dynamic LoggerAdapter that updates context from a Pydantic model.

    Attributes:
        context_obj: A Pydantic model instance providing dynamic context data.
    """

    def __init__(self, logger: Logger, context_obj):
        """Initialise the DynamicLoggerAdapter.

        Args:
            logger (Logger): The base logger to adapt.
            context_obj (LoggerContext): A Pydantic model instance providing dynamic context data.
        """
        self.context_obj = context_obj
        super().__init__(logger, {})

    def process(self, msg, kwargs):
        """Process the logging message and keyword arguments.

        Args:
            msg ([TODO:parameter]): [TODO:description]
            kwargs ([TODO:parameter]): [TODO:description]

        Returns:
            [TODO:return]
        """
        # Get fresh context data each time
        return msg, {**kwargs, "extra": self.context_obj.dict()}


class LoggerContext(BaseModel):
    """Contains dynamic context information for logging.

    Attributes:
        resource_name: Name of the resource being processed.
        resource_type: Resource type, e.g., 'event', 'n2_acquisition', 'raw_adc'.
        pipeline: Name of the processing pipeline.
        site: Site where the processing is occurring, e.g., 'chime', 'kko', 'gbo', 'hco'.
    """

    resource_name: Optional[str] = None
    resource_type: Optional[Literal["event", "n2_acquisition", "raw_adc"]] = None
    pipeline: Optional[
        Literal[
            "baseband-conversion",
            "datatrail-deletion",
            "datatrail-registration",
            "datatrail-reattempt-unregistered",
            "datatrail-replication",
            "l4-trigger",
        ]
    ] = os.getenv(
        "CHIME_LOGGER_PIPELINE_NAME"
    )  # type: ignore[assignment]
    site: Optional[Literal["chime", "kko", "gbo", "hco"]] = None

    @validator("resource_name", "resource_type", pre=True)
    def _normalise_string_field(cls, v, field):
        """Normalise string fields to lowercase.

        Args:
            v: Value to validate.
            field: Field being validated.

        Returns:
            Validated and normalised value.

        Raises:
            ValueError: If the value is not a string.
        """
        if isinstance(v, str):
            return v.lower()
        raise ValueError(f"{field.name} must be a string")

    class Config:
        """Configuration for the LoggerContext model.

        Attributes:
            anystr_strip_whitespace:
            min_anystr_length:
            validate_assignment:
        """

        anystr_strip_whitespace = True
        min_anystr_length = 1
        validate_assignment = True



import atexit
import logging
import logging.config
import queue
import sys
from logging.handlers import QueueHandler, QueueListener

from chime_logger.config import LOGGING_CONFIG


def setup_logging():
    """Configures logging for the CHIME logger package using the LOGGING_CONFIG dictionary.

    - Sets up handlers, formatters, and filters as specified in LOGGING_CONFIG.
    - Starts the QueueListener for the 'queue_handler' if it exists.
    - Registers atexit cleanup to stop the QueueListener on program exit.
    """
    if sys.version_info >= (3, 12):
        logging.config.dictConfig(LOGGING_CONFIG)
        queue_handler = logging.getHandlerByName("queue_handler")
    else:
        LOGGING_CONFIG["handlers"].pop("queue_handler")
        LOGGING_CONFIG["loggers"]["CHIME"]["handlers"] = ["loki"]
        logging.config.dictConfig(LOGGING_CONFIG)
        logger = logging.getLogger("CHIME")
        queue_handler = QueueHandler(queue.Queue(-1))
        logger.addHandler(queue_handler)
        loki_handler = next(
            (h for h in logger.handlers if h.__class__.__name__ == "LokiHandler"),
            None,
        )
        logger.handlers.pop(0)
        listener = QueueListener(
            queue_handler.queue, loki_handler, respect_handler_level=True
        )
        queue_handler.listener = listener

    if queue_handler is not None:
        queue_handler.listener.start()
        atexit.register(queue_handler.listener.stop)



"""This module provides custom logging filters for the CHIME logger.

Classes:
    PipelineFilter: Ensures log records have a 'pipeline' attribute.
    ResourceNameFilter: Ensures log records have an 'resource_name' attribute.
    ResourceTypeFilter: Ensures log records have an 'resource_type' attribute.
    SiteFilter: Ensures log records have an 'site' attribute.
"""

import logging
import os


class PipelineFilter(logging.Filter):
    """Logging filter that ensures each log record has a 'pipeline' attribute.

    If the 'pipeline' attribute is missing, it sets it to the value of the
    'CHIME_LOGGER_PIPELINE_NAME' environment variable, or 'unknown_pipeline'
    if the environment variable is not set.
    """

    def filter(self, record):
        """Ensures that each log record has a 'pipeline' attribute.

        If the 'pipeline' attribute is missing from the log record, this method sets it
        to the value of the 'CHIME_LOGGER_PIPELINE_NAME' environment variable, or
        'unknown_pipeline' if the environment variable is not set.

        Args:
            record (logging.LogRecord): The log record to filter.

        Returns:
            bool: Always returns True to allow the record to be processed.
        """
        # If the record does not have a 'pipeline' attribute, set it to a default value
        if not hasattr(record, "pipeline"):
            record.pipeline = os.getenv(
                "CHIME_LOGGER_PIPELINE_NAME", "unknown_pipeline"
            )
        return True


class ResourceNameFilter(logging.Filter):
    """Logging filter that ensures each log record has an 'resource_name' attribute.

    If the 'resource_name' attribute is missing, it sets it to 'unknown_name'.
    """

    def filter(self, record):
        """Ensures that each log record has an 'resource_name' attribute.

        If the 'resource_name' attribute is missing from the log record, this method sets it
        to 'unknown_name'.

        Args:
            record (logging.LogRecord): The log record to filter.

        Returns:
            bool: Always returns True to allow the record to be processed.
        """
        # If the record does not have an 'resource_name' attribute, set it to a default value
        if not hasattr(record, "resource_name"):
            record.resource_name = "unknown_name"
        return True


class ResourceTypeFilter(logging.Filter):
    """Logging filter that ensures each log record has an 'resource_type' attribute.

    If the 'resource_type' attribute is missing, it sets it to 'unknown_type'.
    """

    def filter(self, record):
        """Ensures that each log record has an 'resource_type' attribute.

        If the 'resource_type' attribute is missing from the log record, this method sets it
        to 'unknown_type'.

        Args:
            record (logging.LogRecord): The log record to filter.

        Returns:
            bool: Always returns True to allow the record to be processed.
        """
        # If the record does not have an 'resource_type' attribute, set it to a default value
        if not hasattr(record, "resource_type"):
            record.resource_type = "unknown_type"
        return True


class SiteFilter(logging.Filter):
    """Logging filter that ensures each log record has an 'site' attribute.

    If the 'site' attribute is missing, it sets it to 'unknown_site'.
    """

    def filter(self, record):
        """Ensures that each log record has an 'site' attribute.

        If the 'site' attribute is missing from the log record, this method sets it
        to 'unknown_site'.

        Args:
            record (logging.LogRecord): The log record to filter.

        Returns:
            bool: Always returns True to allow the record to be processed.
        """
        # If the record does not have an 'site' attribute, set it to a default value
        if not hasattr(record, "site"):
            record.site = "unknown_site"
        return True
